/*
 * Jakelib2 - General purpose C++ library
 * Copyright (C) 2001 Florian Wolff  (florian@donuz.de)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: ZString.h,v 1.3 2002-09-16 18:39:16 gcc3 Exp $
 */

class ZObject;
class ZHashtable;

#undef printf

/**
 * The class <tt>ZString</tt> includes methods for examining 
 * individual characters of the sequence, for comparing strings, for 
 * searching strings, for extracting substrings, and for
 * converting strings to upper or lower case.
 *
 * @package jakelib.lang
 * @version $Id: ZString.h,v 1.3 2002-09-16 18:39:16 gcc3 Exp $
 * @author Florian 'Overflo' Wolff (florian@donuz.de)
 */
class ZString : public ZObject
{
 public:
  /**
   * Use <tt>printf</tt> instead!  Constructs a new string from a <tt>printf</tt>-style format string.
   * The first argument is just a dummy to allow the compiler to
   * distinguish between the different constructors.
   *
   * @param dummy May be any int value
   * @param fstring A <tt>printf</tt>-style format string
   * @param ... Arguments to be inserted as defined by the format string
   * @deprecated
   */
  ZString(int dummy, const char* fstring, ...);
  
  /**
   * Constructs a new string as a copy of a given string.
   *
   * @param str The string to be copied
   */
  ZString(const ZString& str);
  
  /**
   * Constructs a new string from a given array of characters ( <tt>char*</tt> ).
   *
   * @param str The C-style string
   */
  ZString(const char* str);
  
  /**
   * Constructs a new string as a copy of a given string.
   *
   * @param str The string to be copied
   */
  ZString(ZString* str);
  
  /**
   * Constructs a new and empty string.
   */
  ZString();
  
  /**
   * <P> Constructs a new "Null" or "normal" string. A "Null" string may be usefull 
   * if a function returns a string by value and in some cases should return
   * some kind of invalid string.</P>
   * <P> Whether a given string is a "Null" string or not can be determined by
   * <tt>isNull</tt>.
   *
   * @param isNullString If <tt>TRUE</tt> a "Null" string is created, otherwise a normal string is created.
   */
  ZString(boolean isNullString);
  
  /**
   * Guess what: This is just a lousy destructor!
   */
  ~ZString();
  
  /**
   * Returns a pointer to a copy of this string object.
   *
   * @return ZString*
   */
  //      ZString clone();
  
  /**
   * Returns a hashcode for this string. The hashcode for a 
   * <code>String</code> object is computed as
   * <blockquote><pre> 
   * s[0]*31^(n-1) + s[1]*31^(n-2) + ... + s[n-1]
   * </pre></blockquote>
   * using <code>int</code> arithmetic, where <code>s[i]</code> is the 
   * <i>i</i>th character of the string, <code>n</code> is the length of 
   * the string, and <code>^</code> indicates exponentiation. 
   * (The hash value of the empty string is zero.)
   */
  int hashCode();
  
  /**
   * Returns the length of this string.
   *
   * @return The length of the sequence of characters represented by this object.
   */
  int length();
  
  /**
   * Return the string.
   */
  ZString toString();
      
  /**
   * Compares this object to the specified object.
   * The result is <tt>TRUE</tt> if and only if the argument is not
   * <tt>NULL</tt> and is a <tt>ZString</tt> object that contains
   * the same string value as this object.
   *
   * @param obj The object to compare with.
   * @return <tt>TRUE</tt> if the objects are the same; <tt>FALSE</tt> otherwise.
   */
  boolean equals(ZObject* obj);
  
  /**
   * Compares this string to the specified C-style string.
   * The result is <tt>TRUE</tt> if and only if the argument is not
   * <tt>NULL</tt> and contains the same character sequence
   * as this string.
   *
   * @param str The C-style string to compare with.
   * @return <tt>TRUE</tt> if the strings are equal; <tt>FALSE</tt> otherwise.
   */
  boolean equals(const char* str);
  
  /**
   * Compares this string case-insensitively to the specified C-style string.
   * The result is <tt>TRUE</tt> if and only if the argument is not
   * <tt>NULL</tt> and contains the same character sequence
   * as this string.
   *
   * @param str The C-style string to compare with.
   * @return <tt>TRUE</tt> if the strings are equal; <tt>FALSE</tt> otherwise.
   */
  boolean equalsIgnoreCase(const char* str);
  
  /**
   * Returns a copy of this string where some character substitutions
   * have been made. <BR>
   * <TT>&lt;nl&gt;</TT> =&gt; <TT>\n</TT><BR>
   * <TT>&lt;rt&gt;</TT> =&gt; <TT>\r</TT><BR>
   * <TT>\</TT> =&gt; <TT>\\</TT><BR>
   * <TT>&lt;tab&gt;</TT> =&gt; <TT>\t</TT><BR>
   * <TT>"</TT> =&gt; <TT>\"</TT><BR>
   */
  ZString toCppString();
  ZString parseCppString();
  
  /**
   * Removes all preceeding blanks.
   */
  ZString& ltrim();
  
  /**
   * Removes all trailing blanks.
   */
  ZString& rtrim();
  
  /**
   * Removes blanks from both ends of the string.
   */
  ZString& trim();
  
  /**
   * Returns the integer value of the string.
   *
   * @return Integer value
   */
  int toInt();
  
  /**
   * Returns the long integer value of the string.
   *
   * @return Integer value
   */
  long toLong();
  
  /**
   * Returns a pointer to the array of characters in this string.
   *
   * @return Pointer to the internal array of characters.
   */                   
  const char* getChars();
  
  /**
   * Returns a pointer to the array of characters in this string.
   *
   * @param index The start index
   * @return Pointer to the internal array of characters.
   * @throws IllegalArgumentException If index is greater than the length of the string.
   */                   
  const char* getChars(int index);
  
  int indexOf(ZString* str);
  int indexOf(const char* str);
  int indexOf(char c);
  int indexOf(ZString* str, int fromIndex);
  int indexOf(const char* str, int fromIndex);
  int lastIndexOf(ZString* str);
  int lastIndexOf(const char*);
  
  /**
   * Tests if this string starts with the specified prefix.
   *
   * @param   prefix The prefix.
   * @return  <tt>TRUE</tt> if the character sequence represented by the
   *          argument is a prefix of the character sequence represented by
   *          this string; <tt>FALSE</tt> otherwise.      
   *          Note also that <tt>TRUE</tt> will be returned if the 
   *          argument is an empty string or is equal to this 
   *          string object. 
   */
  boolean endsWith(ZString* prefix);
  boolean endsWith(const char* prefix);
  boolean endsWithIgnoreCase(const char* prefix);
  
  /**
   * Tests if this string ends with the specified suffix.
   *
   * @param suffix The suffix.
   * @return  <tt>TRUE</tt> if the character sequence represented by the
   *          argument is a suffix of the character sequence represented by
   *          this string; <tt>FALSE</tt> otherwise.      
   *          Note also that <tt>TRUE</tt> will be returned if the 
   *          argument is an empty string or is equal to this 
   *          string object. 
   */
  boolean startsWith(ZString* suffix);
  boolean startsWith(const char* suffix);
  boolean startsWithIgnoreCase(const char* suffix);
  
  ZString substring(int beginIndex, int endIndex);
  ZString substring(int beginIndex);
  
  char charAt(int index);
  ZString& setChar(int index, char c);

  /**
   * <p> Converts all characters in this string to lower case.
   * <b> Unlike the corresponding Java method
   * (<tt>java.lang.String.toLowerCase</tt>), the string itself is 
   * modified!!</b> </p>
   * <p> If you want the more java-like behaviour use <tt>getLowerCase</tt>.</p>
   *
   * @return A Reference to the string itself.
   */
  ZString& toLowerCase();

  /**
   * <p> Converts all characters in this string to upper case.
   * <b> Unlike the corresponding Java method
   * (<tt>java.lang.String.toUpperCase</tt>), the string itself is 
   * modified!</b> </p>
   * <p> If you want the more java-like behaviour use <tt>getUpperCase</tt>.</p>
   *
   * @return A Reference to the string itself.
   */
  ZString& toUpperCase();
  ZString getLowerCase();
  ZString getUpperCase();
  ZString& replace(char oldChar, char newChar);
  
  //ZString* soundex();
  
  ZString& remove(int beginIndex, int endIndex);
  ZString& remove(int beginIndex);
  
  ZString& append(const ZString* str);
  ZString& append(const char* str);
  ZString& append(char c);
  
  boolean isNull();
  
  ZString& insert(int index, char c);
  //void insert(int index, ZString& str);
  ZString& insert(int index, const char* str);
  
  ZString& clear();
  void ensureCapacity(int minimumCapacity);
  
  ZString& replaceStrings(const char* leftBrace, const char* rightBrace, ZHashtable* hash);
  
  ZString& operator = (const ZString& str);
  ZString& operator = (const char* str);
  ZString& operator += (const ZString& str);
  ZString& operator += (const char* str);
  
  int compareTo(ZObject* o);

  /**
   * @param fstring A <tt>printf</tt>-style format string
   * @param ... Arguments to be inserted as defined by the format string
   */
  ZString& printf(const char * fstring, ...);
  
 protected:
  boolean isNullString;

  /** Contains the length of the string in characters. */
  int len;

  /** Contains the max. capacity of the string in characters. */
  int capacity;

  /** Contains a pointer the the array of charaters containing the string data. */
  char* buffer;

  static int initialCapacity;
  static int increment;

  void init(int initialCapacity);
};


#ifndef HAVE_ITOA
char *itoa (int value, char *string, int radix);

#endif

